/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include "mptkcodec.h"

#define DATA_LEN_MAX 1<<17

bitplane::bitplane( unsigned int set_data_len, double set_greatest_amp, unsigned int max_numBytes, unsigned int numBlocks, dictionary *dict, unsigned int sampleRate, bool set_is_psy ) {

	data_len = set_data_len;
	greatest_amp = set_greatest_amp;
	sig = new unsigned int[data_len];
	newsig = new unsigned int[data_len];
	select = new unsigned int[data_len];
	level = new long unsigned int[data_len];
	bs = new bitstream(max_numBytes);
	s = new slot( numBlocks, dict );
	s->init_mdct();
	mask = NULL;

	is_psy = set_is_psy;

	if (is_psy) {
		s->init_psy( sampleRate );
		mask = new slot( numBlocks, dict );
	}
}

bitplane::~bitplane( void ) {
	delete[]( sig );
	delete[]( newsig );
	delete[]( select );
	delete[]( level );
	delete( bs );
	delete( s );
	if ( mask ) delete( mask );
}

void bitplane::plug_data( double *set_data ) {

	data = set_data;
}

void bitplane::init( unsigned int target_numBytes  ) {

	/* Declarations */
	unsigned int n;

	/* Max value */
	data_max_q = (double) data_max_q_u * pow(2,-DATA_MAX_NBITS) * greatest_amp;

	/* Init arrays */
	for ( n=0 ; n<data_len ; n++ ) {
		sig[n] = 0;
		select[n] = 1;
		level[n] = 2;
		newsig[n] = 0;
	}

	/* Init number of new significant coefficients */
	num_nsc = 0;

	/* Init partially coded coefficients */
	s->reset_data();
	s->interleave_data();

	/* Init bit counter */
	bits_counter_old = 0;
	bits_counter_new = 0;
}

unsigned int bitplane::code( double * set_data, unsigned int target_numBytes ) {

	/* Declarations */
	unsigned int n;
	double amp;

	/* Data */
	data = set_data;

	/* Init bitstream */
	bs->reset();
	bs->set_maxNumBits( target_numBytes*8 );

	/* Quantize and code max value */
	data_max = 0;
	for ( n=0 ; n<data_len ; n++ ) {
		amp = data[n];
		if (amp<0) amp = -amp;
		if (amp>data_max) data_max = amp;
	}
	data_max_q_u = (unsigned int)ceil( pow(2,DATA_MAX_NBITS) * data_max / greatest_amp );
	bs->put_unsigned(data_max_q_u,DATA_MAX_NBITS);

	/* Init bitplane */
	init( target_numBytes );

	/* LOOP */
	while (true) {

		/* Reset new significant coefficients array */
		for ( n=0 ; n<data_len ; n++ ) {
			if ( newsig[n] == 1 ) num_nsc++;
			newsig[n] = 0;
		}

		/* Update the mask */
		if (is_psy) update_mask();

		/* Select next coefficients to be encoded */
		if (is_psy) select_next();

		/* Significance map and sign encoding */
		if ( sigmap_sign_code() ) break;

		/* Refinement pass */
		if ( refine_code() ) break;

		/* Update the level */
		if ( update_level() ) break;
	}

	/* Deinterleave partially coded coefficients */
	s->deinterleave_data();

	/* Flush Bitsream */
	bs->flush();

	/* Return number of bytes written */
	return bs->numBytes;
}

unsigned int bitplane::decode( unsigned int target_numBytes ) {

	/* Declarations */
	unsigned int n;

	/* Init bitstream */
	bs->reset();
	bs->set_maxNumBits( target_numBytes*8 );

	/* Decode max value */
	bs->get_unsigned(&data_max_q_u,DATA_MAX_NBITS);

	/* Init bitplane */
	init( target_numBytes );

	while (true)  {

		/* Reset new significant coefficients array */
		for ( n=0 ; n<data_len ; n++ ) {
			if ( newsig[n] == 1 ) num_nsc++;
			newsig[n] = 0;
		}

		/* Update the mask */
		if (is_psy) update_mask();

		/* Select next coefficients to be decoded */
		if (is_psy) select_next();

		/* Significance map and sign decoding */
		if ( sigmap_sign_decode() ) break;

		/* Refinement pass */
		if ( refine_decode() ) break;

		/* Update the level */
		if ( update_level() ) break;
	}

	/* Deinterleave decoded coefficients */
	s->deinterleave_data();

	/* Return number of bytes */
	return bs->numBytes;
}

int bitplane::sigmap_sign_code( void ) {

	/* Declarations */
	unsigned int runlen[DATA_LEN_MAX],signs[DATA_LEN_MAX];
	unsigned int bp_len, num_ones, num_ones_coded,n, sign, current_pos, last_pos;
	double amp;
	long int num_zeros,k,k2;
	int ret;

	/* Init return */
	ret = 0;

	/* Compute run-lengths and signs */
	bp_len = 0;
	num_ones = 0;
	current_pos = 0;
	last_pos = 0;
	for ( n=0 ; n<data_len ; n++ ) {
		if ( select[n]==1 && sig[n]==0 ) {
			bp_len++;
			current_pos++;
			amp = data[n];
			sign = 1;
			if (amp<0) { amp = -amp; sign = 0; }
			if ( amp*level[n] > data_max_q ) {
				runlen[num_ones] = current_pos - last_pos - 1;
				last_pos = current_pos;
				signs[num_ones] = sign;
				num_ones++;
			}
		}
	}
	runlen[num_ones] = current_pos - last_pos;

	/* Code run-lengths and signs */
	k = K_INIT;
	k2 = 1<<K_INIT;
	num_ones_coded = 0;
	for ( n=0 ; n<num_ones ; n++ ) {
		num_zeros = runlen[n];
		sign = signs[n];
		while ( num_zeros>=k2 ) {
			num_zeros = num_zeros - k2;
			if ( bs->put_bit(1) ) {ret=1;goto End;}
			k++;
			k2 = k2<<1;
		}
		if ( bs->put_bit(0) ) {ret=1;goto End;}
		if ( bs->put_unsigned(num_zeros,k) ) {ret=1;goto End;}
		if (k>0) { k--; k2=k2>>1; }
		if ( bs->put_bit(sign) ) {ret=1;goto End;}
		num_ones_coded++;
	}

	/* Code last number of zeros */
	num_zeros = runlen[num_ones];
	while ( num_zeros>=0 ) {
		num_zeros = num_zeros - k2;
		if ( bs->put_bit(1) ) {ret=1;goto End;}
		k++;
		k2 = k2<<1;
	}

	End:

	/* Set new significant coefficients */
	current_pos = 0;
	last_pos = 0;
	num_ones = 0;
	for ( n=0 ; n<data_len ; n++ ) {
		if (num_ones==num_ones_coded) break;
		if ( select[n]==1 && sig[n]==0 ) {
			current_pos++;
			if ( current_pos-last_pos-1==runlen[num_ones] ) {
				new_coef( n, signs[num_ones] );
				last_pos = current_pos;
				num_ones++;
				sig[n] = 1;
				newsig[n] = 1;
			}
		}
	}

	return ret;
}

int bitplane::sigmap_sign_decode( void ) {

	/* Declarations */
	unsigned int runlen[DATA_LEN_MAX],signs[DATA_LEN_MAX];
	unsigned int bp_len, num_ones, n, current_pos, last_pos, cum, u;
	long int num_zeros,k,k2;
	int ret,bit;

	/* Init return */
	ret = 0;

	/* Compute bitplane length */
	bp_len = 0;
	for ( n=0 ; n<data_len ; n++ ) if ( select[n]==1 && sig[n]==0 ) bp_len++;

	/* Decode run-lengths and signs */
	cum = 0;
	num_ones = 0;
	k = K_INIT;
	k2 = 1<<K_INIT;
	while ( true ) {
		num_zeros = 0;
		while (true) {
			if ( bs->get_bit(&bit) ) {ret = 1; goto End;}
			if (bit==0) break;
			num_zeros = num_zeros + k2;
			cum = cum + k2;
			if ( cum>bp_len) goto End;
			k++;
			k2 = k2<<1;
		}
		if ( bs->get_unsigned(&u,k) ) {ret = 1; goto End;}
		num_zeros = num_zeros + u;
		cum = cum + u + 1;
		if (k>0) { k--; k2=k2>>1; }
		if ( bs->get_bit(&bit) ) {ret = 1; goto End;}
		runlen[num_ones] = num_zeros;
		signs[num_ones] = (unsigned int)bit;
		num_ones++;
	}
	End:
	runlen[num_ones] = data_len;

	/* Set new significant coefficients */
	current_pos = 0;
	last_pos = 0;
	num_ones = 0;
	for ( n=0 ; n<data_len ; n++ ) {
		if ( select[n]==1 && sig[n]==0 ) {
			current_pos++;
			if ( current_pos-last_pos-1==runlen[num_ones] ) {
				new_coef( n, signs[num_ones] );
				last_pos = current_pos;
				num_ones++;
				sig[n] = 1;
				newsig[n] = 1;
			}
		}
	}

	return ret;
}

int bitplane::refine_code( void ) {

	/* Declarations */
	unsigned int n,ref_bit;
	double amp;

	/* Loop */
	for ( n=0 ; n<data_len ; n++ ) {
		if ( select[n]==1 && sig[n]==1 && newsig[n]==0 ) {
			amp = data[n];
			if (amp<0) amp = -amp;
			ref_bit = ((unsigned int)floor(amp*level[n]/data_max_q))%2;
			if ( bs->put_bit(ref_bit) ) return 1;
			refine_coef( n, ref_bit );
		}
	}

	return 0;
}

int bitplane::refine_decode( void ) {

	/* Declarations */
	unsigned int n;
	int ref_bit;

	/* Loop */
	for ( n=0 ; n<data_len ; n++ ) {
		if ( select[n]==1 && sig[n]==1 && newsig[n]==0 ) {
			if ( bs->get_bit(&ref_bit) ) return 1;
			refine_coef( n, ref_bit );
		}
	}

	return 0;
}

int bitplane::update_level( void ) {

	unsigned int n;
	long unsigned int level_max = (1<<LEVEL_MAX);
	for ( n=0 ; n<data_len ; n++ ) {
		if ( select[n]==1 ) {
			level[n] = level[n] << 1;
		}
		if (level[n]>=level_max) return 1;
	}
	return 0;
}

void bitplane::new_coef( unsigned int n, unsigned int sign ) {

	s->dataI[n] = 1.5 * data_max_q / level[n];
	if (sign==0) s->dataI[n] = -s->dataI[n];
}

void bitplane::refine_coef( unsigned int n, unsigned int ref_bit ) {

	unsigned int sign;
	double amp,tmp;

	amp = s->dataI[n];
	sign = 1;
	if (amp<0) { amp = -amp; sign = 0; }
	tmp = data_max_q / (2*level[n]);
	if (ref_bit==0) tmp = -tmp;
	s->dataI[n] =  amp + tmp;
	if (sign==0) s->dataI[n] = -s->dataI[n];
}

void bitplane::select_next( void ) {

	/* Declarations */
	double gap[DATA_LEN_MAX];
	unsigned int n,count;
	double gap_mean;

	/* Compute gap between current level and masking threshold */
	gap_mean = 0;
	count = 0;
	for ( n=0 ; n<data_len ; n++ ) {
		if (mask->dataI[n]==0) printf("WARNING: MASK NULL \n");
		gap[n] = -log2( sqrt( mask->dataI[n] ) / data_max_q ) - log2(level[n]);
		gap_mean += gap[n];
		count++;
	}
	gap_mean /= count;

	/* Select coefficients above gap mean */
	for ( n=0 ; n<data_len ; n++ ) {
		if ( gap[n]>=gap_mean+GAP_ADJUST ) {
			select[n] = 1;
		} else {
			select[n] = 0;
		}
	}
}

void bitplane::update_mask( void ) {

	bits_counter_new = bs->numBits;
	if ( (bits_counter_new-bits_counter_old>UPDATE_PSY)||bits_counter_old==0 ) {

		bits_counter_old = bits_counter_new;

		/* Deinterleave data */
		s->deinterleave_data();
	
		/* Compute the mask */
		s->compute_mask( mask );
	
		/* Interleave the mask */
		mask->interleave_data();
	}
}
