/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include "mptkcodec.h"

book_coder::book_coder( MP_Book_c *book, unsigned int set_numSamples, unsigned int mode ) {

	/* Declarations */
	unsigned int n,center,index;
	MP_Mdct_Atom_c *atom;

	/* Number of samples */
	numSamples = set_numSamples;

	/* Sample rate */
	sampleRate = book->sampleRate;

	/* Dictionary */
	numBlocks = (unsigned int)round(log2(WIN_LEN_MAX/WIN_LEN_MIN)+1);
	dict = new dictionary[numBlocks];
	for ( n=0 ; n<numBlocks ; n++ ) {
		dict[n].windowLen = WIN_LEN_MIN << n;
		dict[n].windowType = WIN_TYPE;
		dict[n].windowOption = WIN_OPTIONAL;
	}

	/* Slot parameters: length, position of first slot, number of slots */
	slotLen = dict[numBlocks-1].windowLen >> 1;
	firstSlotPos = slotLen;
	numSlots = numSamples/slotLen + 1;

	/* Init book array */
	bookA = new MP_Book_c*[numSlots];
	for ( n=0 ; n<numSlots ; n++ ) bookA[n] = MP_Book_c::init();

	/* Fill book array */
	firstNonZeroSlot = numSamples;
	lastNonZeroSlot = 0;
	for ( n=0 ; n<book->numAtoms ; n++ ) {
		center = book->atom[n]->support[0].pos + (book->atom[n]->support[0].len >> 1);
		index = (center-firstSlotPos)/slotLen;
		if (index<0||index>numSlots-1) printf("Error: atom too close to the edges\n");
		if (index>lastNonZeroSlot) lastNonZeroSlot = index;
		if (index<firstNonZeroSlot) firstNonZeroSlot = index;
		atom = MP_Mdct_Atom_c::init( 1, book->atom[n]->get_field(MP_WIN_TYPE_PROP,0), book->atom[n]->get_field(MP_WIN_OPT_PROP,0) );
		atom->support[0].len = book->atom[n]->support[0].len;
		atom->support[0].pos = book->atom[n]->support[0].pos;
		atom->numSamples = book->atom[n]->numSamples;
		atom->freq = book->atom[n]->get_field(MP_FREQ_PROP,0);
		atom->amp[0] = book->atom[n]->amp[0];
		bookA[index]->append( atom );
	}
	numSlots = lastNonZeroSlot - firstNonZeroSlot + 1;

	/* Init slot */
	s = new slot(numBlocks,dict);

	/* Mode */
	switch (mode) {
		case 0:
			is_psy = false;
			break;
		case 1:
			is_psy = true;
			break;
		default:
			is_psy = false;
			printf("Unknown mode\n");
			break;
	}

	/* Init bitplane */
	bp = NULL;

	/* Init signal */
	signal = NULL;
}

book_coder::~book_coder( void ) {

	/* Declarations */
	unsigned int n;
	numSlots = numSamples/slotLen + 1;

	/* Clean memory */
	delete[]( dict );
        for ( n=0 ; n<numSlots ; n++ ) delete( bookA[n] );
	delete[]( bookA );
	delete( s );
	if ( bp ) delete( bp );
	if ( signal ) delete( signal );
}

void book_coder::init_mdct( void ) {

	s->init_mdct();
}

void book_coder::init_signal( void ) {

	signal = MP_Signal_c::init( 1, numSamples, sampleRate );
}

void book_coder::init_bitplane( void ) {

	unsigned int max_numBytes = (unsigned int)( ((double)slotLen * (double)MAX_BITRATE * (double)1000) / ((double)8*(double)sampleRate) ) + 1;
	bp = new bitplane(numBlocks*slotLen,sqrt(slotLen*2),max_numBytes,numBlocks,dict,sampleRate, is_psy);
}

unsigned int book_coder::write_header( FILE *output ) {

	unsigned int numBytes = 0;

	/* Write number of samples */
	fwrite(&numSamples,sizeof(unsigned int),1,output);
	numBytes += sizeof(unsigned int);

	/* Write sample rate */
	fwrite(&sampleRate,sizeof(unsigned int),1,output);
	numBytes += sizeof(unsigned int);

	/* Write first and last non-zeros Slot */
	fwrite(&firstNonZeroSlot,sizeof(unsigned int),1,output);
	numBytes += sizeof(unsigned int);
	fwrite(&lastNonZeroSlot,sizeof(unsigned int),1,output);
	numBytes += sizeof(unsigned int);

	/* Write mode */
	fwrite(&is_psy,sizeof(bool),1,output);
	numBytes += sizeof(bool);

	/* Return number of bytes */
	return numBytes;
}

void book_coder::load_slot( unsigned int idx_slot ) {

	/* Load atoms and fill coefficients array */
	s->load_atoms(bookA[idx_slot],idx_slot*slotLen+firstSlotPos );
}

void book_coder::interleave_slot( void ) {

	/* Interleave data */
	s->interleave_data();
}

unsigned int book_coder::code_slot( unsigned int bitrate ) {

	unsigned int max_numBytes, numBytes;

	/* Code vector of interleaved coefficients with target bitrate */
	max_numBytes = (unsigned int)( ((double)slotLen * (double)bitrate * (double)1000) / ((double)8*(double)sampleRate) ) + 1;
	numBytes = bp->code( s->dataI, max_numBytes );

	/* Return number of bytes */
	return numBytes;
}

void book_coder::synth_coded_slot( unsigned int idx_slot ) {

	/* Declarations */
	unsigned int n;

	/* Synthesize slot */
	if ( signal!=NULL ) {
		bp->s->synthesize( );
		for ( n=0 ; n<(slotLen<<1) ; n++ ) signal->storage[idx_slot*slotLen+firstSlotPos-(slotLen>>1)+n] += bp->s->signal[n];
	}
}

unsigned int book_coder::write_slot( FILE *output ) {

	/* Write data */
	fwrite(bp->bs->bytes,sizeof(unsigned char),bp->bs->numBytes,output);

	/* Return number of bytes */
	return bp->bs->numBytes;
}

void book_coder::synth_slot( unsigned int idx_slot ) {

	/* Declarations */
	unsigned int n;

	/* Synthesize slot */
	if ( signal!=NULL ) {
		s->synthesize( );
		for ( n=0 ; n<(slotLen<<1) ; n++ ) signal->storage[idx_slot*slotLen+firstSlotPos-(slotLen>>1)+n] += s->signal[n];
	}
}
