/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include "mptkcodec.h"

book_decoder::book_decoder( unsigned int mode ) {

	/* Init objects */
	book = NULL;
	dict = NULL;
	bp = NULL;
	signal = NULL;

	/* Mode */
	switch (mode) {
		case 0:
			is_book = false;
			is_wav = true;
			break;
		case 1:
			is_book = true;
			is_wav = false;
			break;
		default:
			is_book = false;
			is_wav = true;
			printf("Unknown mode\n");
			break;
	}
}

book_decoder::~book_decoder( void ) {

	/* Clean memory */
	if ( book ) delete( book );
	if ( dict ) delete[]( dict );
	if ( bp ) delete( bp );
	if ( signal ) delete( signal );
}

unsigned int book_decoder::read_header( FILE *input ) {

	/* Declarations */
	unsigned int numBytes = 0;

	/* Read number of samples */
	fread(&numSamples,sizeof(unsigned int),1,input);
	numBytes += sizeof(unsigned int);

	/* Read sample rate */
	fread(&sampleRate,sizeof(unsigned int),1,input);
	numBytes += sizeof(unsigned int);

	/* Read first and last non-zeros Slot */
	fread(&firstNonZeroSlot,sizeof(unsigned int),1,input);
	numBytes += sizeof(unsigned int);
	fread(&lastNonZeroSlot,sizeof(unsigned int),1,input);
	numBytes += sizeof(unsigned int);

	/* Read Mode */
	fread(&is_psy,sizeof(bool),1,input);
	numBytes += sizeof(bool);

	/* Init decoder */
	init_decoder();

	return numBytes;
}

void book_decoder::init_decoder( void ) {

	/* Declarations */
	unsigned int max_numBytes,n;

	/* Dictionary */
	numBlocks = (unsigned int)round(log2(WIN_LEN_MAX/WIN_LEN_MIN)+1);
	dict = new dictionary[numBlocks];
	for ( n=0 ; n<numBlocks ; n++ ) {
		dict[n].windowLen = WIN_LEN_MIN << n;
		dict[n].windowType = WIN_TYPE;
		dict[n].windowOption = WIN_OPTIONAL;
	}

	/* Slot parameters: length, position of first slot, number of slots */
	slotLen = dict[numBlocks-1].windowLen >> 1;
	firstSlotPos = slotLen;
	numSlots = lastNonZeroSlot - firstNonZeroSlot + 1;

	/* Init book */
	if ( is_book ) {
		book = MP_Book_c::init();
		book->sampleRate = sampleRate;
	}

	/* Init bitplane */
	max_numBytes = (unsigned int)( ((double)slotLen * (double)MAX_BITRATE * (double)1000) / ((double)8*(double)sampleRate) ) + 1;
	bp = new bitplane(numBlocks*slotLen,sqrt(slotLen*2),max_numBytes,numBlocks,dict,sampleRate, is_psy);

	/* Init signal */
	if ( is_wav ) signal = MP_Signal_c::init( 1, numSamples, sampleRate );
}

unsigned int book_decoder::read_slot( FILE *input, unsigned int bitrate, unsigned int maxNumBytes ) {

	/* Declarations */
	unsigned int numBytes;

	/* Read data */
	numBytes = (unsigned int)( ((double)slotLen * (double)bitrate) / ((double)8*(double)sampleRate) ) + 1;
	if (maxNumBytes<numBytes) {
		numBytes = maxNumBytes;
	}
	fread(bp->bs->bytes,sizeof(unsigned char),numBytes,input);

	return numBytes;
}

void book_decoder::decode_slot( unsigned int idxSlot, unsigned int numBytes ) {

	/* Declarations */
	unsigned int n;

	/* Decode data */
	bp->decode(numBytes);

	/* Save atoms */
	if ( is_book ) bp->s->save_atoms(book,idxSlot*slotLen+firstSlotPos );

	/* Synthesise slot */
	if ( is_wav ) {
		bp->s->synthesize( );
		for ( n=0 ; n<(slotLen<<1) ; n++ ) signal->storage[idxSlot*slotLen+firstSlotPos-(slotLen>>1)+n] += bp->s->signal[n];
	}
}
