/***************************************************************************
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>
#include "mptkcodec.h"

psy::psy( unsigned int set_len, unsigned int set_sampleRate ) {

	/* Declarations */
	unsigned int n,m;
	double tmp;

	/* Frame length */
	len = set_len;
	input_len = len<<1;

	/* Sample Rate */
	sampleRate = set_sampleRate;

	/* Partition width (in bark) */
	part_width = PART_WIDTH;

	/* Spectral flatness maximum value */
	sfm_db_max = SFM_DB_MAX;

	/* Frequency bins in Hz */
	freq_hz = new double[len];
	for ( n=0 ; n<len ; n++ ) freq_hz[n] = (n+0.5)*sampleRate/(2*len);

	/* Frequency bins in Bark */
	freq_bark = new double[len];
	for ( n=0 ; n<len ; n++ ) {
		freq_bark[n] = hz2bark( freq_hz[n] );
	}

	/* Number of partitions */
	tmp = 0;
	nb_part = 0;
	for ( n=1 ; n<len ; n++ ) {
		tmp += freq_bark[n] - freq_bark[n-1];
		if (tmp>part_width) {
			nb_part++;
			tmp = 0;
		}
	}
	nb_part++;

	/* Lowest and highest frequency line in each partition */
	low = new unsigned int[nb_part];
	high = new unsigned int[nb_part];
	tmp = 0;
	nb_part = 0;
	low[0] = 0;
	for ( n=1 ; n<len ; n++ ) {
		tmp += freq_bark[n] - freq_bark[n-1];
		if (tmp>part_width) {
			high[nb_part] = n-1;
			low[nb_part+1] = n;
			nb_part++;
			tmp = 0;
		}
	}
	nb_part++;
	high[nb_part-1] = len-1;

	/* Number of frequency bins in each partition */
	nb_bins = new unsigned int[nb_part];
	for ( n=0 ; n<nb_part ; n++ ) nb_bins[n] = high[n] - low[n] + 1;

	/* Partition of each frequency bin */
	part = new unsigned int[len];
	for ( n=0 ; n<nb_part ; n++ ) {
		for ( m=low[n] ; m<=high[n] ; m++ ) part[m] = n;
	}

	/* Median bark and hz value */
	median_bark = new double[nb_part];
	median_hz = new double[nb_part];
	for ( n=0 ; n<nb_part ; n++ ) {
		median_hz[n] = (freq_hz[low[n]]+freq_hz[high[n]])/2;
		median_bark[n] = hz2bark(median_hz[n]);
	}

	/* Absolute threshold of hearing */
	ath_db = new double[nb_part];
	ath_lin = new double[nb_part];
	for ( n=0 ; n<nb_part ; n++ ) {
		ath_db[n] = ath( median_hz[n] );
		ath_lin[n] = pow(10,ath_db[n]/10);
	}

	//for ( n=0 ; n<nb_part ; n++ ) printf("test %u %u %u %u %f %f\n",n,nb_bins[n],low[n],high[n],median_bark[n],ath_db[n]);

	/* Normalisation coefficients */
	rnorm = new double[nb_part];
	for ( n=0 ; n<nb_part ; n++ ) {
		tmp = 0;
		for ( m=0 ; m<nb_part ; m++ ) {
			tmp += sprdngf(median_bark[m],median_bark[n]);
		}
		rnorm[n] = 1 / tmp;
	}

	/* Noise masking tone, tone masking noise */
	tmn = TMN;
	nmt = NMT;

	/* Init arrays */
	input = new double[input_len];
	r = new double[len];
	e = new double[nb_part];
	ecb = new double[nb_part];
	nb = new double[nb_part];
	smr = new double[nb_part];
	mdct_e = new double[nb_part];
	snr = new double[nb_part];
	bc = new double[nb_part];	
}

psy::~psy( void ) {
	delete[]( freq_hz );
	delete[]( freq_bark );
	delete[]( low );
	delete[]( high );
	delete[]( nb_bins );
	delete[]( median_bark );
	delete[]( median_hz );
	delete[]( ath_db );
	delete[]( ath_lin );
	delete[]( rnorm );
	delete[]( input );
	delete[]( r );
	delete[]( e );
	delete[]( ecb );
	delete[]( nb );
	delete[]( smr );
	delete[]( mdct_e );
	delete[]( part );
	delete[]( snr );
	delete[]( bc );
}

int psy::compute( mdct *m, double *signal, double *mask ) {

	/* Declarations */
	unsigned int n,l,zeros;
	double gmean_db,amean_db,sfm_db;
	double coef;

	/* O) Test if the signal is zeros,  ATH if zeros */
	zeros = 1;
	for ( n=0 ; n<input_len ; n++ ) {
		if (signal[n]!=0) {
			zeros = 0;
			break;
		}
	}
	if (zeros==1) {
		for (n=0;n<nb_part;n++) {
			for (l=low[n];l<=high[n];l++) mask[l] = ath_lin[n];
		}
		return 0;
	}

	/* 1) Input samples */
	for ( n=0 ; n<input_len ; n++ ) {
		input[n] = signal[n];
	}

	/* 2) MCLT */
	m->compute_mclt( input );
	for ( n=0 ; n<len ; n++ ) {
		r[n] = sqrt(m->mcltE[n] / 2);
	}

	/* 3) Energy in each partition */
	for ( n=0 ; n<nb_part ; n++ ) {
		e[n] = 0;
		for ( l=low[n] ; l<=high[n] ; l++ ) e[n] += r[l]*r[l];
	}

	/* 4) Convolution of the energy with the spreading function */
	for ( n=0 ; n<nb_part ; n++ ) {
		ecb[n] = 0;
		for ( l=0 ; l<nb_part ; l++ ) {
			ecb[n] += e[l]*sprdngf(median_bark[l],median_bark[n]);
		}
	}
	for ( n=0 ; n<nb_part ; n++ ) ecb[n] *= rnorm[n];

	/* 5) Tonality index (based on the spectral flatness measure) */
	gmean_db = 0;
	amean_db = 0;
	for ( n=0 ; n<len ; n++ ) {
		gmean_db += 10 * log10( r[n]*r[n] ) / len;
		amean_db += r[n]*r[n] / len;
	}
	amean_db = 10*log10( amean_db );
	sfm_db = gmean_db - amean_db;
	if (sfm_db/ SFM_DB_MAX>1) {
		tb = 1;
	} else {
		tb = sfm_db/ SFM_DB_MAX;
	}

	/* 6) Required SNR in each partition */
	for ( n=0 ; n<nb_part ; n++ ) snr[n] = tb*TMN + (1-tb)*NMT;

	/* 7) Power ratio */
	for ( n=0 ; n<nb_part ; n++ ) bc[n] = pow( 10, -snr[n]/10);

	/* 8) Actual energy threshold */
	for ( n=0 ; n<nb_part ; n++ ) nb[n] = bc[n] * ecb[n];

	/* 9) Threshold in quiet */
	for ( n=0 ; n<nb_part ; n++ ) {
		if (nb[n]<(ath_lin[n]*(high[n]-low[n]+ 1))) nb[n] = (ath_lin[n]*(high[n]-low[n]+ 1));
	}

	/* 10) Signal to Mask Ratio */
	for ( n=0 ; n<nb_part ; n++ ) {
		smr[n] = e[n] / nb[n];
	}

	/* 11) Energy of MDCT in each band */
	for ( n=0 ; n<nb_part ; n++ ) {
		mdct_e[n] = 0;
		for ( l=low[n] ; l<=high[n] ; l++ ) {
			coef = m->mcltRe[l];
			mdct_e[n] += coef*coef;
		}
		mdct_e[n] /= (high[n] - low[n] + 1);
		if (mdct_e[n]==0) mdct_e[n]=1e-16;
	}

	/* 12) Masking threshold */
	for ( n=0 ; n<nb_part ; n++ ) {
		for ( l=low[n] ; l<=high[n] ; l++ ) {
			mask[l] = mdct_e[n] / smr[n];
		}
	}

	return 0;
}

double psy::hz2bark( double f ) {
	return 13*atan(0.00076*f) + 3.5*atan(pow(f/7500,2));
}

double psy::ath( double f ) {
	if (f<ATH_F_LOW) f = ATH_F_LOW;
	if (f>ATH_F_HIGH) f = ATH_F_HIGH;
	f = f / 1000;
	double t = 3.64*pow(f,-0.8)-6.5*exp(-0.6*pow(f-3.3,2))+pow(10,-3)*pow(f,4) + ATH_ADJUST;
	return t;
}

double psy::sprdngf( double i, double j ) {

	double tmpx,tmpy,tmpz,r;

	if (j >= i ) {
		tmpx = 3 * (j - i);
	} else {
		tmpx = 1.5 * (j - i);
	}
	tmpz = 8 * (pow(tmpx - 0.5,2) - 2 * (tmpx - 0.5));
	if ( tmpz > 0 ) {
		tmpz = 0;
	}
	tmpy = 15.811389 + 7.5 * (tmpx + 0.474) - 17.5 * (pow((1 + pow(tmpx + 0.474,2)),0.5));
	if ( tmpy < -100 ) { 
		r = 0;
	} else {
		r = pow(10,((tmpz + tmpy) / 10));
	}
	return r;
}
